@@ -82,7 +82,7 @@
 #include "GriddedTableDef.h"
 #include "UngriddedTableDef.h"
 #include "Function.h"
-#include "SignalDef.h"
+#include "SgnlDef.h"
 #include "CheckData.h"
 #include "Author.h"
 #include "Reference.h"
@@ -718,6 +718,9 @@ namespace janus {
      * \return A reference to the list of BreakpointDef instances is returned.
      */
     BreakpointDefList& getBreakpointDef() { return breakpointDef_;}
+    const BreakpointDefList& getBreakpointDef() const { return breakpointDef_;}
+
+    const BreakpointDef& getBreakpointDef( const dstoute::aString& bpId) const;
 
     /**
      * Within the DOM, a \em griddedTableDef contains points arranged in an
